//  Copyright (c) 2004 by Mark F. Adams 
//  $Id: agg_smooth_gs.C,v 1.8 2004/11/23 18:04:01 adams Exp $
//  Author:             Mark F. Adams
//  Filename:           amg_smooth_gs.C
//
#include "prometheus.hh"

#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::SmoothP_gs" 
/* PromAggGrid::SmoothP_gs ****************************************************
 * 
 *  - replace next->restriction: P := (I - (gauss-seidel)A)P 
 *  - collective on 'wcomm'
 *
 *  INPUT:SmoothP_gs
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *     - PETSc error code 
 */ 
int PromAggGrid::SmoothP_gs() const
{
  int   ierr,pp,np;                assert(stiffness_ != NULL);
  const PromMatrix * const AA = stiffness_;  assert(AA!=NULL);

  perf_mon_.EventBegin( PromContext::prom_logs_[LOG_MAPS] );
  if( isActive() ) {
    //
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS1]);
    ierr = FormDAP_gs( AA );	                 CHKERRQ(ierr);
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS1]);

    // output P
    if( comm_.np() == -1 ){ 
      char name[25];
      sprintf( name, "P_1_level_%d", getLevel() );
      FILE *file = fopen( name, "w" ); 
      for( int xx = 0 ; xx < nNodes_ ; xx++ ){
	PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
	int geqi, geqj, fndf = fine->getNDF(), ii, jj; 
	const PromParentArr *parI = fine->parents_; 
	for( pp = 0, np = fine->nParents() ; pp < np ; pp++ ){
	  int cndf = parI->data_[pp]->getNDF();
	  assert(parI->data_[pp]->DAP_ != NULL); // no!!
	  for(ii = 0, geqi = fine->geq() ; ii < fndf ; ii++, geqi++){
	    const shpfloat * const pp0 = &parI->data_[pp]->DAP_[ii*cndf]; 
	    for(jj=0, geqj=parI->data_[pp]->geq() ; jj<cndf ; jj++, geqj++){
	      if(fabs(pp0[jj]) > 1.e-12 )
		fprintf(file,"%7d%7d%15.7e\n",geqi,geqj,pp0[jj]);
	    }
	  }
	}
      }
      fclose(file);
    }
  }
  else {
    perf_mon_.EventBegin( PromContext::prom_logs_[LOG_MIS1] );
    perf_mon_.EventEnd( PromContext::prom_logs_[LOG_MIS1] );
  }
  perf_mon_.Flops(1); // make sure this is not in symbolic
  perf_mon_.EventEnd( PromContext::prom_logs_[LOG_MAPS] );

  return 0;
}

//#define SKIP_NEAR_BOUND
#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::FormDAP_gs" 
/* PromAggGrid::FormDAP_gs ************************************************
 *
 *  - form D^-1 A P: "parents->DAP_" <- D^-1 A "parents->P0_"  
 *
 *  INPUT:
 *   - AA: stiffness matrix
 *
 *  SIDE EFFECTS:
 *   - allocs DAP
 *
 *  RETURN: 
 *     - PETSc error code 
 */
int PromAggGrid::FormDAP_gs( const PromMatrix * const AA ) const
{
  const int  my0=my0Nd(); PromParent *pari; 
  const int fact=AA->isComplex() ? 2 : 1,fact2=fact*fact;
  int        ierr,jj,ii,kk,tt,xx,ncols,pp,flops=0,adjacs[PROM_MAXNIEGH],did,todo,proc,hits;
  double    *Ai[PROM_MAXNIEGH];  assert(AA != NULL);  shpfloat *pDAP; 
  xx=sizeof(shpfloat); pp=sizeof(MPI_Request);  ii=sizeof(int);
  const int fisz = xx/ii, reqsz = (pp/xx + ((pp%xx == 0) ? 0 : 1)) * fisz;
  MPI_Status  status;  PromList<PromNode*> *list;
  PromList<MPI_Request*> smessages;
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromList<int> procs(10); const int tag = getNewTag(); 
  if(AA->isSymmetricMatrix())SETERRQ(-1,"FormDAP called with symmetic matrix");

  // push P0 into DAP, count todos, set proc lists
  for( xx = todo = 0 ; xx < nNodes_ ; xx++ ){
    PromNode *finei = nodes_[xx];
    pari = finei->parents_->data_[0]; 
    const int nparI = finei->nParents(), fndfi = finei->getNDF();
    const int cndfj = pari->getNDF(), sz = cndfj*fndfi;
#if defined(SKIP_NEAR_BOUND)
    int isfx = 0;
#endif
    // alloc and zero if necc.
    if( (pDAP=pari->DAP_) == NULL ) {
      int bsz = sizeof(shpfloat)*sz; assert(bsz>0);
      ierr = PetscMalloc( bsz, &pDAP );  CHKERRQ(ierr); 
      pari->DAP_ = pDAP; PetscMemzero( pDAP, bsz ); 
      assert( pari->getPsz() == sz );
    }
    // copy up
    if( cndfj == 1 ) pDAP[0] = pari->getScalarP0();
    else {
      const shpfloat* const P0 = pari->P0_; assert(P0 != NULL);
      for(ii=0;ii<sz;ii++) pDAP[ii] = P0[ii];
    }
    if( xx < nLocalNd_ ) {
      // row of fine matrix
      ierr = AA->GetLocalNodeRow(xx, PROM_MAXNIEGH, ncols, adjacs); 
      CHKERRQ(ierr);
      // for whole row, get parent array and ndf
      for( int col = 0; col < ncols ; col++ ) {
	const int gidj=adjacs[col],lidj=gidj-my0,isloc=(lidj>=0&&lidj<nLocalNd_);
	PromNode *finej =  isloc ? nodes_[lidj] : ghosts_.find(gidj+1);
	PromParentArr_extra *gpararr = NULL;
	if( finej != NULL ) {
#if defined(SKIP_NEAR_BOUND)
	  if( finej->isFixed() ) isfx++;
#endif
	}
	else { // extra parents from regularization/penalty
	  gpararr = gid_extraParents_->find(gidj+1); assert(0);
	  assert(gpararr!=NULL); assert(gpararr->getGID() == gidj);
#if defined(SKIP_NEAR_BOUND)
	  if( gpararr->isFixed() ) isfx++;
#endif 
	}

	// check 'todo'
	if( !isloc ) {
	  if( gpararr == NULL ) {
	    if( !finej->isMarked() ) { todo++; finej->mark(); }
	  }
	  else if( !gpararr->isMarked() ) { todo++; gpararr->mark(); }
	  // update procs
	  ierr = GetProc( gidj, proc );   CHKERRQ(ierr);
	  if( !procs.find(proc+1) ) procs.addTail(proc+1);// tiny linear search
	}
      }
    }

    // collect bound to send
    while( !procs.isEmpty() ) {
      proc = procs.removeHead() - 1;  
      if( (list=proc_list.find(proc+1)) == NULL ) {
	list = new PromList<PromNode*> (128); CHKNEW(list,1);
	proc_list.add( list, proc+1 ); 
      }
      list->addTail( finei );
    }    
#if defined(SKIP_NEAR_BOUND)
    if( isfx || finei->isFixed() ) 
#else
    if( finei->isFixed() ) 
#endif
      {
	ierr = finei->SetSkipSmooth();       CHKERRQ(ierr);
      }

    // debug
    for(pp=1;pp<nparI;pp++)assert(finei->parents_->data_[pp]->DAP_==NULL);

    // skip smoothing to keep npar < PROM_MAXPAR_SOFT -- invarient!!!
    if( nparI > PROM_MAXPAR_SOFT || finei->skipSmooth() ) { // 
      for( pp = 1 ; pp < nparI ; pp++ ){
	PromParent *par = finei->parents_->data_[pp];
	if( par->DAP_ != NULL ) PetscFree( par->DAP_ ); 
	par->DAP_ = NULL; 
      }
      ierr = finei->SetnParents( parBank_, 1 );  CHKERRQ(ierr);
      ierr = finei->SetSkipSmooth();  CHKERRQ(ierr);
    }

  }
  assert(nNodes_==ghosts_.getCount()+nLocalNd_);

  hits = 0;
  // DAP_i := -sum( DAP_ij * DAP_j ) / A_ii
  for( xx = nLocalNd_-1 ; xx >= 0 ; xx-- ){
    PromNode *finei = nodes_[xx];
    if( finei->skipSmooth() ) continue;
    PromParentArr *parIarr = finei->parents_; 
    PromParent **parIdata = parIarr->data_; 
    int nparI = finei->nParents(), gidi = finei->index_.global();
    int fndfj, fndfi = finei->getNDF();

    /* zero out rows */
    pari = parIdata[0]; kk = pari->getNDF()*fndfi;
    assert(pari->DAP_ != NULL);
    for(ii=0,pDAP=pari->DAP_;ii<kk;ii++) pDAP[ii] = 0.;

    // row of fine matrix
    ierr = AA->GetLocalNodeRow(xx, PROM_MAXNIEGH, ncols, adjacs, Ai); 
    CHKERRQ(ierr);
    // for whole row, get parent array and ndf
    for( int col = 0; col < ncols ; col++ ) {
      const int gidj=adjacs[col],lidj=gidj-my0,isloc=(lidj>=0&&lidj<nLocalNd_);
      if( gidj == gidi ) { did = col; continue; }
      PromNode *finej =  isloc ? nodes_[lidj] : ghosts_.find(gidj+1);
      const PromParentArr *parJarr = NULL; PromParentArr_extra *gpararr = NULL;
      if( finej != NULL ) {
	fndfj = finej->getNDF(); parJarr = finej->parents_;
      }
      else { // extra parents from regularization/penalty
	gpararr = gid_extraParents_->find(gidj+1); assert(0); // not used
	assert(gpararr!=NULL); assert(gpararr->getGID() == gidj);
	parJarr = gpararr; fndfj = gpararr->getNDF();
      }
      const double *const Aij = Ai[col]; 

      // apply DAP_iJ -= A_ij * DAP_jJ 
      for( int ppp = 0 ; ppp < parJarr->getSize() ; ppp++ ){
	const PromParent * const parj = parJarr->data_[ppp]; // DAP_jJ
	if( parj->DAP_ == NULL ){ assert(ppp>0); continue; }
	const int pargidJ = parj->global(), cndfj = parj->getNDF(); 
	assert(cndfj>0);
	// find pari (DAP_iJ)
	pari = NULL;
	for( pp = 0 ; pp < nparI && pari == NULL; pp++ ){
	  if(parIdata[pp]->global() == pargidJ) pari = parIdata[pp];
	}

	if( pari == NULL ) ;
	else {
	  // alloc and zero if necc.
	  if( (pDAP=pari->DAP_) == NULL ) {
	    //assert(kkk==0);
	    int sz = cndfj*fndfi, bsz = sizeof(shpfloat)*sz; assert(bsz>0);
	    ierr = PetscMalloc( bsz, &pDAP );  CHKERRQ(ierr); 
	    pari->DAP_ = pDAP; PetscMemzero( pDAP, bsz ); 
	    //pDAP[sz] = PROM_BIG_NUM;
	    if( pari->getPsz() == 0 ) {
	      ierr = pari->SetPsz_hard( sz ); CHKERRQ(ierr);
	    }
	    else if( pari->getPsz() != sz ){
	      SETERRQ4(1,"locid=%d, cndfj=%d par->psz()=%d, sz=%d",
		       xx,cndfj,pari->getPsz(),sz);
	    }
	  }
	  //assert(pDAP[cndfj*fndfi] == PROM_BIG_NUM);
	  hits++; // count cells
	  // do DAP_iJ -= A_ij * P_jJ
	  if( cndfj == 1 ) *pDAP -= *parj->DAP_ * *Aij;
	  else{
	    const shpfloat* const P0 = parj->DAP_; assert(P0 != NULL);
	    const double *aijT = Aij; // transpose
	    int iii,ij,jjj,ii;
	    // buffer := A_ij^T * P_jJ 
	    if( cndfj == 6 && fndfi%3 == 0 ) { // unroll
	      for( ii = 0, iii = 0, ij = 0 ; ii < fndfj ; ii++, iii += 6 ){
		shpfloat t0 = P0[iii],  t1 = P0[iii+1],t2 = P0[iii+2];
		shpfloat t3 = P0[iii+3],t4 = P0[iii+4],t5 = P0[iii+5];
		for(jj=jjj = 0 ; jj < fndfi; jj += 3,ij += 3*fact2,jjj += 18){
		  shpfloat *pv = &pDAP[jjj];
		  double aij0=aijT[ij],aij1=aijT[ij+fact],aij2=aijT[ij+2*fact];
		  pv[0] -= aij0 * t0;	    pv[1] -= aij0 * t1;
		  pv[2] -= aij0 * t2;	    pv[3] -= aij0 * t3;
		  pv[4] -= aij0 * t4;	    pv[5] -= aij0 * t5;
		  pv[6] -= aij1 * t0;	    pv[7] -= aij1 * t1;
		  pv[8] -= aij1 * t2;	    pv[9] -= aij1 * t3;
		  pv[10] -= aij1 * t4;	    pv[11] -= aij1 * t5;
		  pv[12] -= aij2 * t0;	    pv[13] -= aij2 * t1;
		  pv[14] -= aij2 * t2;	    pv[15] -= aij2 * t3;
		  pv[16] -= aij2 * t4;	    pv[17] -= aij2 * t5;
		}
	      }
	    }
	    else {
	      const double *paijt = &aijT[0]; shpfloat *pv = &pDAP[0];
	      for( ii = 0 ; ii < fndfi ; ii++, paijt++ ) {
		const shpfloat * pp0 = &P0[0], *p0;  
		for( jj = 0, kk ; jj < cndfj ; jj++, pv++, pp0++ ) {
		  const double *pa = paijt + ii*fact; p0 = pp0; 
		  for(kk=0;kk<fndfj;kk++,p0+=cndfj,pa+=fndfi*fact2) {
		    *pv -= *pa**p0;
		  }
		}
	      }
	    }
	  }
	  flops += fndfj*fndfi*cndfj; 
	  //assert(pDAP[cndfj*fndfi] == PROM_BIG_NUM);
	}
      }
    }
    assert( adjacs[did] == gidi );
    const double * const aijT = Ai[did];
#if defined(PROM_USE_NODAL_SMAGG)
    // v <- D^-1 v 
    double Ainv[4*PROM_MAXNDF*PROM_MAXNDF],tarr[4*PROM_MAXNDF*PROM_MAXNDF];
    int nnz = PromContext::TInvert(aijT,Ainv,fndfi,fndfi*fact); assert(nnz==0);
    // smooth
    for( int pp = 0 ; pp < finei->nParents() ; pp++ ) {
      PromParent *par = finei->parents_->data_[pp];
      if( par->DAP_ != NULL ) {
	const int cndf = par->getPsz()/fndfi; assert(cndf>0);
	shpfloat *pB, *pv = Ainv;           assert(cndf==par->getNDF());
	for( ii = 0, jj ; ii < fndfi ; ii++, pv += fndfi ) {
	  for( jj = 0, pB = par->DAP_ ; jj < cndf ; jj++, pB++ ) {
	    double t = 0.0; 
	    for( kk = 0 ; kk < fndfi ; kk++ ) t += pv[kk] * pB[cndf*kk];
	    tarr[cndf*ii + jj] = t;
	  }
	}
	for( ii = 0, pv = tarr, pB = par->DAP_ ; ii < fndfi ; ii++ ){
	  for( jj = 0 ; jj < cndf ; jj++ ) *pB++ = *pv++;
	}
	flops += fndfi*fndfi*(fndfi+cndf);
      }
    }
#else
    for( pp = 0 ; pp < finei->nParents() ; pp++ ){
      PromParent *par = finei->parents_->data_[pp];
      const int cndf = par->getNDF(); assert(cndf>0);
      shpfloat *pDAP = par->DAP_;
      if( pDAP != NULL ) {
	for( ii = 0, tt = 0 ; ii < fndfi ; ii++ ){
	  shpfloat dtemp = aijT[ii*fndfi+ii];
	  if( dtemp < 1.e-100 ) {
	    PetscPrintf(PETSC_COMM_SELF,
			"\t[%d]%s warning diag[%d,ndf %d]=%e\n",
			comm_.mype(),__FUNCT__,xx,ii,dtemp);
	    dtemp = 1.;
	  }
	  dtemp = 1. / dtemp;
	  for( kk = 0 ; kk < cndf ; kk++, tt++ ) pDAP[tt] *= dtemp;
	}
	flops += cndf*fndfi;
      }
    }
#endif
  }
  perf_mon_.Flops(2*flops); flops = 0;
#if defined(PETSC_USE_BOPT_g)
  ierr = PetscTrValid(__LINE__, __FUNCT__, __FILE__, __SDIR__); CHKERRQ(ierr);
#endif
  if(options_.verbose_>1 && comm_.mype()==0)
    PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s hits=%d\n",comm_.mype(),__FUNCT__,hits);

   // clear up unmarked ghost's pars (not touched but covered by local agg.)
  for( xx = nLocalNd_ ; xx < nNodes_ ; xx++ ){
    PromNode *finej = nodes_[xx];
    if( !finej->isMarked() ) {
      if( finej->nParents() > 1 ) {
	ierr = finej->SetnParents( parBank_, 1 );  CHKERRQ(ierr);
      }
      assert(finej->parents_->data_[0]->DAP_ != NULL);
    }
  }
  if( gid_extraParents_ != NULL ) {
    TablePos tpos = gid_extraParents_->getHeadPosition(); 
    while( tpos ) { 
      PromParentArr_extra *parr = gid_extraParents_->getNext( tpos );
      if( !parr->isMarked() ) {
	if( parr->getSize() > 1 ) {
	  ierr = parr->SetSize( parBank_, 1 );  CHKERRQ(ierr);
	}
	assert(parr->data_[0]->DAP_ != NULL);
      }
    }
  }

  // send back parents
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    Position pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0;
    while( pos ) {
      PromNode *finei = list->getNext( pos ); 
      const int fndf = finei->getNDF();
      tt += 2; // gid, npar
      for( jj = 0 ; jj < finei->nParents() ; jj++ ) {
	PromParent *par = finei->parents_->data_[jj];
	tt += par->getNDF()*fndf*fisz + 2; // psz, gidJ, ...
      }
    }
    MPI_Request *sreq;
    ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );     
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      PromNode *finei = list->removeHead(); 
      const int fndf = finei->getNDF();
      *pb++ = finei->index_.global(); int *npar = pb++, pp;
      for( jj = 0, pp = 0 ; jj < finei->nParents() ; jj++ ) {
	PromParent *par = finei->parents_->data_[jj];
	pDAP = par->DAP_;  
	int sz = par->getNDF()*fndf;
	if(!par->DAP_)PetscPrintf(PETSC_COMM_SELF,
				  "\t\t[%d]%s %d) pp=%d, pDAP=%p sz=%d\n",
				  comm_.mype(),__FUNCT__,finei->index_.global(),
				  pp,par->DAP_,sz);
	else PetscPrintf(PETSC_COMM_SELF,"[%d]",comm_.mype());
	*pb++ = sz; *pb++ = par->global();
	shpfloat *fpb = (shpfloat*)pb;
	if(pDAP != NULL)for( ii = 0 ; ii < sz ; ii++ ) *fpb++ = *pDAP++;
	else for( ii = 0 ; ii < sz ; ii++ ) *fpb++ = .0;
	pb = (int*)fpb;
	pp++;
      }
      *npar = pp;
    }
    assert(tt == pb-buff);
    // send(tag) *< gid, npar, npar*< psz, gidJ, DAP[]> >
    MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list; 
  }

  // recieve parents
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt );
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) < gidj, npar, npar*<psz, jidJ, DAP[]> >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
    while( pb-buff < tt ) {
      const int gidj = *pb++, npRev = *pb++;
      PromNode *finej = ghosts_.find(gidj+1); 
      PromParentArr *pararr = NULL;  int fndf;
      if( finej != NULL ){
	assert(finej->isMarked());
	pararr = finej->parents_;  assert(pararr!=NULL); finej->mark(FALSE);
	fndf = finej->getNDF(); 
      }
      else {
	assert(gid_extraParents_!=NULL);
	PromParentArr_extra *gpararr = gid_extraParents_->find(gidj+1);
	assert( gpararr != NULL );
	pararr = gpararr; assert( gpararr != NULL );
	assert(gpararr->isMarked()); gpararr->mark( FALSE );
	fndf = gpararr->getNDF();
      }
      todo--;  assert(todo>=0);
      const int parSzOrig = pararr->getSize();
      if( npRev != parSzOrig ){
	ierr = pararr->SetSize( parBank_, npRev );  CHKERRQ(ierr);
      }
      for( jj = 0 ; jj < npRev ; jj++ ) {
	PromParent *par = pararr->data_[jj];
	const int sz = *pb++, gidJ = *pb++; 
	if( par->global(ierr) != gidJ ) { // recover
assert(0); // ????
	  int cndf = sz/fndf; assert(sz%fndf==0); assert(par->global(ierr)<0);
	  par->SetNDF( cndf ); 
	  par->setGlobal( gidJ ); //par->setProc(p); par->setGEQ(g);
	}
	assert( par->global() == gidJ );
	if( (pDAP=par->DAP_) == NULL && sz > 0 ) { 
	  ierr = PetscMalloc(sz*sizeof(shpfloat),&pDAP); CHKERRQ(ierr); 
	  par->DAP_ = pDAP;
	  if( par->getPsz() == 0 ) par->SetPsz_hard( sz );
	  else assert( par->getPsz() == sz);
	}
	//else assert(par->DAP_ == NULL);
	shpfloat *fpb = (shpfloat*)pb; 
	for( ii = 0 ; ii < sz ; ii++ ) pDAP[ii] = fpb[ii]; // copy in
	pb = (int*)(fpb + sz);
      } 
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }

  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &ii, &status ); assert(!ii);
  
  return 0;
} 
